// ----------------------------------------------------------------------------
//
//  Copyright (C) 2015-2017 Fons Adriaensen <fons@linuxaudio.org>
//    
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ----------------------------------------------------------------------------


#include <string.h>
#include <math.h>
#include "sn3dpan3.h"


#define DEG2RAD 0.01745329f


static void sn3dpangains (float az, float el, float *g)
{
    float t, x1, y1, z1, x2, y2, z2, c2, s2;

    // SN3D panning gains

    az *= DEG2RAD;
    el *= DEG2RAD; 
    t = cosf (el);
    g [0] = 1.0f;
    g [3] = x1 = cosf (az) * t;
    g [1] = y1 = sinf (-az) * t;
    g [2] = z1 = sinf (el);

    x2 = x1 * x1;
    y2 = y1 * y1;
    z2 = z1 * z1;
    c2 = x2 - y2;
    s2 = 2 * x1 * y1;
    t = 0.866025f;
    g [8] = t * c2;
    g [4] = t * s2;
    t *= 2 * z1;
    g [7] = t * x1;
    g [5] = t * y1;
    g [6] = 1.5f * z2 - 0.5f;

    t = 0.790569f;
    g [15] = t * (x2 - 3 * y2) * x1;
    g [ 9] = t * (3 * x2 - y2) * y1;
    t = 1.936492f * z1;
    g [14] = t * c2;
    g [10] = t * s2;
    t = 0.612372f * (5.0f * z2 - 1.0f);
    g [13] = t * x1; 
    g [11] = t * y1; 
    g [12] = (2.5f * z2 - 1.5f) * z1;
}


//-----------------------------------------------------------------------------------
// Mono input panner.


const LADSPA_PortDescriptor Ladspa_Sn3dpan3::portdescr [NPORT] = 
{
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL
};


const char * const Ladspa_Sn3dpan3::portnames [NPORT] = 
{
    "in",
    "out.0", "out.1", "out.2", "out.3", "out.4", "out.5", "out.6", "out.7", 
    "out.8", "out.9", "out.10", "out.11", "out.12", "out.13", "out.14", "out.15", 
    "Elev",
    "Azim"
};


const LADSPA_PortRangeHint Ladspa_Sn3dpan3::porthints [NPORT] = 
{
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, -90, 90 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, -180, 180 }
};


LADSPA_Handle Ladspa_Sn3dpan3::create (const struct _LADSPA_Descriptor *desc, unsigned long rate)
{
    return new Ladspa_Sn3dpan3 (rate);
}


void Ladspa_Sn3dpan3::setport (unsigned long port, LADSPA_Data *data)
{
    _port [port] = data;
}


void Ladspa_Sn3dpan3::active (bool act)
{
    if (act) memset (_G1, 0, 16 * sizeof (float));
}



void Ladspa_Sn3dpan3::runproc (unsigned long len, bool add)
{
    unsigned int  i, j, k, n;
    float  G [16], D [16], buff [BSIZE];
    float  g, d, *out;

    memcpy (G, _G1, 16 * sizeof (float));
    sn3dpangains (_port [CTL_AZ][0], _port [CTL_EL][0], _G1);

    for (i = 1; i < 16; i++)
    {
	d = _G1 [i] - G [i];
	if (fabsf (d) < 1e-3f) d = 0;
	D [i] = d / len;
    }

    for (k = 0; k < len; k += n)
    {
	n = len - k;
	if (n > BSIZE) n = BSIZE;

	memcpy (buff, _port [INP] + k, n * sizeof (float));
	memcpy (_port [OUT] + k, buff, n * sizeof (float));

	for (i = 1; i < 16; i++)
	{
	    g = G [i];
	    d = D [i];
	    out = _port [OUT + i] + k;
	    if (d == 0.0f)
	    {
		for (j = 0; j < n; j++)
		{
		    out [j] = g * buff [j];
		}
	    }
	    else
	    {
		for (j = 0; j < n; j++)
		{
		    g += d;
		    out [j] = g * buff [j];
		}
		G [i] = g;
	    }
	}
    }
}


//-----------------------------------------------------------------------------------
// Stereo input panner.


const LADSPA_PortDescriptor Ladspa_Sn3dpan3s::portdescr [NPORT] = 
{
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL,
    LADSPA_PORT_INPUT  | LADSPA_PORT_CONTROL
};


const char * const Ladspa_Sn3dpan3s::portnames [NPORT] = 
{
    "in.L", "in.R",
    "out.0", "out.1", "out.2", "out.3", "out.4", "out.5", "out.6", "out.7", 
    "out.8", "out.9", "out.10", "out.11", "out.12", "out.13", "out.14", "out.15", 
    "Elev",
    "Azim",
    "Width"
};


const LADSPA_PortRangeHint Ladspa_Sn3dpan3s::porthints [NPORT] = 
{
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { 0, 0, 0 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, -90, 90 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, -180, 180 },
    { LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0, 0, 180 }
};


LADSPA_Handle Ladspa_Sn3dpan3s::create (const struct _LADSPA_Descriptor *desc, unsigned long rate)
{
    return new Ladspa_Sn3dpan3s (rate);
}


void Ladspa_Sn3dpan3s::setport (unsigned long port, LADSPA_Data *data)
{
    _port [port] = data;
}


void Ladspa_Sn3dpan3s::active (bool act)
{
    if (act)
    {
	memset (_GL1, 0, 16 * sizeof (float));
	memset (_GR1, 0, 16 * sizeof (float));
    }
}



void Ladspa_Sn3dpan3s::runproc (unsigned long len, bool add)
{
    unsigned int  i, j, k, n;
    float  az, el, hw, gl, gr, dl, dr;
    float  DL [16], DR [16], GL [16], GR [16], buffL [BSIZE], buffR [BSIZE];
    float  *out;
    
    az = _port [CTL_AZ][0];
    el = _port [CTL_EL][0];
    hw = _port [CTL_SW][0] / 2;

    memcpy (GL, _GL1, 16 * sizeof (float));
    memcpy (GR, _GR1, 16 * sizeof (float));
    sn3dpangains (az - hw, el, _GL1);
    sn3dpangains (az + hw, el, _GR1);

    for (i = 1; i < 16; i++)
    {
	dl = (_GL1 [i] - GL [i]);
	dr = (_GR1 [i] - GR [i]);
	if (fabsf (dl) < 1e-3f) dl = 0;
	if (fabsf (dr) < 1e-3f) dr = 0;
	DL [i] = dl / len;
	DR [i] = dr / len;
    }

    for (k = n = 0; k < len; k += n)
    {
	n = len - k;
	if (n > BSIZE) n = BSIZE;

	memcpy (buffL, _port [INP + 0] + k, n * sizeof (float));
	memcpy (buffR, _port [INP + 1] + k, n * sizeof (float));
	out = _port [OUT] + k;
        for (j = 0; j < n; j++)
        {
	    out [j] = buffL [j] + buffR [j];
        }

        for (i = 1; i < 16; i++)
        {
	    gl = GL [i];
	    gr = GR [i];
	    dl = DL [i];
	    dr = DR [i];
	    out = _port [OUT + i] + k;
            if ((dl == 0.0f) && (dr == 0.0f))
   	    {
	        for (j = 0; j < n; j++)
	        {
	            out [j] = gl * buffL [j] + gr * buffR [j];
  	        }
	    }
	    else
	    {
		for (j = 0; j < n; j++)
		{
		    gl += dl;
		    gr += dr;
		    out [j] = gl * buffL [j] + gr * buffR [j];
		}
	        GL [i] = gl;
  	        GR [i] = gr;
	    }
	}
    }
}


//-----------------------------------------------------------------------------------
